@node rpcgen Programming Guide
@chapter @code{rpcgen} Programming Guide
@iftex
@cindex @code{rpcgen} programming guide
@end iftex
@ifinfo
@cindex rpcgen programming guide
@end ifinfo

This chapter documents the Sun RPC Compiler, the @code{rpcgen}.

@menu
* The rpcgen Protocol Compiler::
* Converting Local Procedures into Remote Procedures::
* Generating XDR Routines::
* The C-Preprocessor::
* rpcgen Programming Notes::
* RPC Language::
@end menu

@node The rpcgen Protocol Compiler
@section The @code{rpcgen} Protocol Compiler
@iftex
@cindex The @code{rpcgen} protocol compiler
@cindex @code{rpcgen}, introduction
@end iftex
@ifinfo
@cindex The rpcgen protocol compiler
@cindex rpcgen, introduction
@end ifinfo

The details of programming applications to use Remote Procedure Calls
can be overwhelming.  Perhaps most daunting is the writing of the XDR
routines necessary to convert procedure arguments and results into their
network format and vice-versa.

Fortunately, @code{rpcgen} exists to help programmers write RPC
applications simply and directly.  @code{rpcgen} does most of the dirty
work, allowing programmers to debug the main features of their
application, instead of requiring them to spend most of their time
debugging their network interface code.

@code{rpcgen} is a compiler.  It accepts a remote program interface
definition written in a language, called RPC Language, which is similar
to C.  It produces a C language output which includes stub versions of
the client routines, a server skeleton, XDR filter routines for both
parameters and results, and a header file that contains common
definitions. The client stubs interface with the RPC library and
effectively hide the network from their callers.  The server stub
similarly hides the network from the server procedures that are to be
invoked by remote clients.  @code{rpcgen}'s output files can be compiled
and linked in the usual way.  The developer writes server procedures ---
in any language that observes Sun calling conventions --- and links them
with the server skeleton produced by @code{rpcgen} to get an executable
server program.  To use a remote program, a programmer writes an
ordinary main program that makes local procedure calls to the client
stubs produced by @code{rpcgen}.  Linking this program with
@code{rpcgen}'s stubs creates an executable program.  (At present the
main program must be written in C).  @code{rpcgen} options can be used
to suppress stub generation and to specify the transport to be used by
the server stub.

Like all compilers, @code{rpcgen} reduces development time that would
otherwise be spent coding and debugging low-level routines.  All
compilers, including @code{rpcgen}, do this at a small cost in
efficiency and flexibility.  However, many compilers allow escape
hatches for programmers to mix low-level code with high-level code.
@code{rpcgen} is no exception.  In speed-critical applications,
hand-written routines can be linked with the @code{rpcgen} output
without any difficulty.  Also, one may proceed by using @code{rpcgen}
output as a starting point, and then rewriting it as necessary.  (If you
need a discussion of RPC programming without @code{rpcgen}, @pxref{Remote
Procedure Call Programming Guide}).

@node Converting Local Procedures into Remote Procedures
@section Converting Local Procedures into Remote Procedures
@cindex Local procedures
@cindex Remote procedures

Assume an application that runs on a single machine, one which we want
to convert to run over the network.  Here we will demonstrate such a
conversion by way of a simple example --- a program that prints a
message to the console:

@example
/*
 * @i{printmsg.c: print a message on the console}
 */
#include <stdio.h>

main(argc, argv)
        int argc;
        char *argv[];
@{
        char *message;

        if (argc < 2) @{
                fprintf(stderr, "usage: %s <message>\n", argv[0]);
                exit(1);
        @}
        message = argv[1];

        if (!printmessage(message)) @{
                fprintf(stderr, "%s: couldn't print your message\n",
                        argv[0]);
                exit(1);
        @} 
        printf("Message Delivered!\n");
        exit(0);
@}
/*
 * @i{Print a message to the console.}
 * @i{Return a boolean indicating whether the message was actually printed.}
 */
printmessage(msg)
        char *msg;
@{
        FILE *f;

        f = fopen("/dev/console", "w");
        if (f == NULL) @{
                return (0);
        @}
        fprintf(f, "%s\n", msg);
        fclose(f);
        return(1);
@}
@end example


And then, of course:

@example
@cartouche
@b{example%} cc printmsg.c -o printmsg
@b{example%} printmsg "Hello, there."
Message delivered!
@b{example%}
@end cartouche
@end example


If @code{printmessage()} was turned into a remote procedure, then it
could be called from anywhere in the network.  Ideally, one would just
like to stick a keyword like @code{remote} in front of a procedure to
turn it into a remote procedure.  Unfortunately, we have to live within
the constraints of the C language, since it existed long before RPC did.
But even without language support, it's not very difficult to make a
procedure remote.

In general, it's necessary to figure out what the types are for all
procedure inputs and outputs.  In this case, we have a procedure
@code{printmessage()} which takes a string as input, and returns an
integer as output.  Knowing this, we can write a protocol specification
in RPC language that describes the remote version of
@code{printmessage()}.  Here it is:

@example
/*
 * @i{msg.x: Remote message printing protocol}
 */
program MESSAGEPROG @{
        version MESSAGEVERS @{
                int PRINTMESSAGE(string) = 1;
        @} = 1;
@} = 99;
@end example


Remote procedures are part of remote programs, so we actually declared
an entire remote program here which contains the single procedure
@code{PRINTMESSAGE}.  This procedure was declared to be in version 1 of
the remote program.  No null procedure (procedure 0) is necessary
because @code{rpcgen} generates it automatically.

Notice that everything is declared with all capital letters.  This is
not required, but is a good convention to follow.

Notice also that the argument type is @code{string} and not @code{char
*}.  This is because a @code{char *} in C is ambiguous.  Programmers
usually intend it to mean a null-terminated string of characters, but it
could also represent a pointer to a single character or a pointer to an
array of characters.  In RPC language, a null-terminated string is
unambiguously called a @code{string}.

There are just two more things to write.  First, there is the remote
procedure itself.  Here's the definition of a remote procedure to
implement the @code{PRINTMESSAGE} procedure we declared above:

@example
/*
 * @i{msg_proc.c: implementation of the remote procedure "printmessage"}
 */
#include <stdio.h>
#include <rpc/rpc.h>    /* @i{always needed}  */
#include "msg.h"        /* @i{need this too: msg.h will be generated by rpcgen} */

/*
 * @i{Remote verson of "printmessage"}
 */
int *
printmessage_1(msg)
        char **msg;
@{
        static int result;  /* @i{must be static!} */
        FILE *f;

        f = fopen("/dev/console", "w");
        if (f == NULL) @{
                result = 0;
                return (&result);
        @}
        fprintf(f, "%s\n", *msg);
        fclose(f);
        result = 1;
        return (&result);
@}
@end example


Notice here that the declaration of the remote procedure
@code{printmessage_1()} differs from that of the local procedure
@code{printmessage()} in three ways:

@enumerate
@item
It takes a pointer to a string instead of a string itself.  This
is true of all  remote procedures:  they always take pointers to  their
arguments rather than the arguments themselves.
@item
It returns a pointer to an  integer instead of  an integer itself. This is
also generally true of remote procedures: they always return a pointer
to their results.
@item
It has an @samp{_1} appended to its name.  In general, all remote
procedures called by @code{rpcgen} are named by the following rule: the
name in the program definition (here @code{PRINTMESSAGE}) is converted
to all lower-case letters, an underbar (@samp{_}) is appended to it, and
finally the version number (here @code{1}) is appended.
@end enumerate

The last thing to do is declare the main client program that will call
the remote procedure. Here it is:

@example
/*
 * @i{rprintmsg.c: remote version of "printmsg.c"}
 */
#include <stdio.h>
#include <rpc/rpc.h>     /* @i{always needed}  */
#include "msg.h"         /* @i{need this too: msg.h will be generated by rpcgen} */

main(argc, argv)
        int argc;
        char *argv[];
@{
        CLIENT *cl;
        int *result;
        char *server;
        char *message;

        if (argc < 3) @{
                fprintf(stderr, "usage: %s host message\n", argv[0]);
                exit(1);
        @}

        /*
         * @i{Save values of command line arguments}
         */
        server = argv[1];
        message = argv[2];

        /*
         * @i{Create client "handle" used for calling @i{MESSAGEPROG} on the}
         * @i{server designated on the command line. We tell the RPC package}
         * @i{to use the "tcp" protocol when contacting the server.}
         */
        cl = clnt_create(server, MESSAGEPROG, MESSAGEVERS, "tcp");
        if (cl == NULL) @{
                /*
                 * @i{Couldn't establish connection with server.}
                 * @i{Print error message and die.}
                 */
                clnt_pcreateerror(server);
                exit(1);
        @}

        /*
         * @i{Call the remote procedure "printmessage" on the server}
         */
        result = printmessage_1(&message, cl);
        if (result == NULL) @{
                /*
                 * @i{An error occurred while calling the server.}
                 * @i{Print error message and die.}
                 */
                clnt_perror(cl, server);
                exit(1);
        @}

        /*
         * @i{Okay, we successfully called the remote procedure.}
         */
        if (*result == 0) @{
                /*
                 * @i{Server was unable to print our message.}
                 * @i{Print error message and die.}
                 */
                fprintf(stderr, "%s: %s couldn't print your message\n", 
                        argv[0], server);       
                exit(1);
        @} 

        /*
         * @i{The message got printed on the server's console}
         */
        printf("Message delivered to %s!\n", server);
@}
@end example

There are two things to note here:

@enumerate
@item
@cindex Client handle, used by @code{rpcgen}
First a @dfn{client handle} is created using the RPC library routine
@code{clnt_create()}.  This client handle will be passed to the stub
routines which call the remote procedure.
@item
The remote procedure @code{printmessage_1()} is called exactly the same
way as it is declared in @file{msg_proc.c} except for the inserted
client handle as the second argument.
@end enumerate

Here's how to put all of the pieces together:

@example
@cartouche
@b{example%} rpcgen msg.x
@b{example%} cc rprintmsg.c msg_clnt.c -o rprintmsg
@b{example%} cc msg_proc.c msg_svc.c -o msg_server
@end cartouche
@end example

Two programs were compiled here: the client program @code{rprintmsg} and
the server program @code{msg_server}.  Before doing this though,
@code{rpcgen} was used to fill in the missing pieces.

Here is what @code{rpcgen} did with the input file @file{msg.x}:

@enumerate
@item
It created a header file called @file{msg.h} that contained
@code{#define}'s for @code{MESSAGEPROG}, @code{MESSAGEVERS} and
@code{PRINTMESSAGE} for use in the other modules.
@item
It created client @samp{stub} routines in the @file{msg_clnt.c} file.
In this case there is only one, the @code{printmessage_1()} that was
referred to from the @code{printmsg} client program.  The name of the
output file for client stub routines is always formed in this way: if
the name of the input file is @file{FOO.x}, the client stubs output file
is called @file{FOO_clnt.c}.
@item
It created the server program which calls @code{printmessage_1()} in
@file{msg_proc.c}.  This server program is named @file{msg_svc.c}.  The
rule for naming the server output file is similar to the previous one:
for an input file called @file{FOO.x}, the output server file is named
@file{FOO_svc.c}.
@end enumerate

Now we're ready to have some fun.  First, copy the server to a remote
machine and run it.  For this example, the machine is called @samp{moon}.
Server processes are run in the background, because they never exit.

@example
@cartouche
@b{moon%} msg_server &
@end cartouche
@end example

Then on our local machine (@samp{sun}) we can print a message on
@samp{moon}s console.

@example
@cartouche
@b{sun%} printmsg moon "Hello, moon."
@end cartouche
@end example

The message will get printed to @samp{moon}s console.  You can print a
message on anybody's console (including your own) with this program if
you are able to copy the server to their machine and run it.

@node Generating XDR Routines
@section Generating XDR Routines
@cindex Generating XDR routines
@cindex XDR routines, generating

The previous example only demonstrated the automatic generation of
client and server RPC code.  @code{rpcgen} may also be used to generate
XDR routines, that is, the routines necessary to convert local data
structures into network format and vice-versa.  This example presents a
complete RPC service --- a remote directory listing service, which uses
@code{rpcgen} not only to generate stub routines, but also to generate
the XDR routines.  Here is the protocol description file:

@example
/*
 * @i{dir.x: Remote directory listing protocol}
 */
const MAXNAMELEN = 255;         /* @i{maximum length of a directory entry} */

typedef string nametype<MAXNAMELEN>;    /* @i{a directory entry} */

typedef struct namenode *namelist;              /* @i{a link in the listing} */

/*
 * @i{A node in the directory listing}
 */
struct namenode @{
        nametype name;          /* @i{name of directory entry} */
        namelist next;          /* @i{next entry} */
@};

/*
 * @i{The result of a READDIR operation.}
 */
union readdir_res switch (int errno) @{
case 0:
        namelist list;  /* @i{no error: return directory listing} */
default:
        void;           /* @i{error occurred: nothing else to return} */
@};

/*
 * @i{The directory program definition}
 */
program DIRPROG @{
        version DIRVERS @{
                readdir_res
                READDIR(nametype) = 1;
        @} = 1;
@} = 76;
@end example

@strong{Note:} @emph{Types (like @code{readdir_res} in the example
above) can be defined using the @code{struct}, @code{union} and
@code{enum} keywords, but those keywords should not be used in
subsequent declarations of variables of those types.  For example, if
you define a union @code{foo}, you should declare using only @code{foo}
and not @code{union foo}.  In fact, @code{rpcgen} compiles RPC unions
into C structures and it is an error to declare them using the
@code{union} keyword.}

Running @code{rpcgen} on @file{dir.x} creates four output files.  Three
are the same as before: header file, client stub routines and server
skeleton.  The fourth are the XDR routines necessary for converting the
data types we declared into XDR format and vice-versa.  These are output
in the file @file{dir_xdr.c}.

Here is the implementation of the @code{READDIR} procedure.

@example
/*
 * @i{dir_proc.c: remote readdir implementation}
 */
#include <rpc/rpc.h>
#include <sys/dir.h>
#include "dir.h"

extern int errno;
extern char *malloc();
extern char *strdup();

readdir_res *
readdir_1(dirname)
        nametype *dirname;
@{
        DIR *dirp;
        struct direct *d;
        namelist nl;
        namelist *nlp;
        static readdir_res res; /* @i{must be static}! */

        /*
         * @i{Open directory}
         */
        dirp = opendir(*dirname);
        if (dirp == NULL) @{
                res.errno = errno;
                return (&res);
        @}

        /*
         * @i{Free previous result}
         */
        xdr_free(xdr_readdir_res, &res);

        /*
         * @i{Collect directory entries.}
         * @i{Memory allocated here will be freed by @code{xdr_free()}}
         * @i{next time @code{readdir_1()} is called}
         */
        nlp = &res.readdir_res_u.list;
        while (d = readdir(dirp)) @{
                nl = *nlp = (namenode *) malloc(sizeof(namenode));
                nl->name = strdup(d->d_name);
                nlp = &nl->next;
        @}
        *nlp = NULL;

        /*
         * @i{Return the result}
         */
        res.errno = 0;
        closedir(dirp);
        return (&res);
@}
@end example

Finally, there is the client side program to call the server:

@example
/*
 * @i{rls.c: Remote directory listing client}
 */
#include <stdio.h>
#include <rpc/rpc.h>    /* @i{always need this} */
#include "dir.h"                /* @i{will be generated by rpcgen} */

extern int errno;

main(argc, argv)
        int argc;
        char *argv[];
@{
        CLIENT *cl;
        char *server;
        char *dir;
        readdir_res *result;
        namelist nl;


        if (argc != 3) @{
                fprintf(stderr, "usage: %s host directory\n", 
                  argv[0]);
                exit(1);
        @}

        /*
         * @i{Remember what our command line arguments refer to}
         */
        server = argv[1];
        dir = argv[2];

        /*
         * @i{Create client "handle" used for calling @i{MESSAGEPROG} on the}
         * @i{server designated on the command line. We tell the RPC package}
         * @i{to use the "tcp" protocol when contacting the server.}
         */
        cl = clnt_create(server, DIRPROG, DIRVERS, "tcp");
        if (cl == NULL) @{
                /*
                 * @i{Couldn't establish connection with server.}
                 * @i{Print error message and die.}
                 */
                clnt_pcreateerror(server);
                exit(1);
        @}

        /*
         * @i{Call the remote procedure @i{readdir} on the server}
         */
        result = readdir_1(&dir, cl);
        if (result == NULL) @{
                /*
                 * @i{An error occurred while calling the server.}
                 * @i{Print error message and die.}
                 */
                clnt_perror(cl, server);
                exit(1);
        @}

        /*
         * @i{Okay, we successfully called the remote procedure.}
         */
        if (result->errno != 0) @{
                /*
                 * @i{A remote system error occurred.}
                 * @i{Print error message and die.}
                 */
                errno = result->errno;
                perror(dir);
                exit(1);
        @}

        /*
         * @i{Successfully got a directory listing.}
         * @i{Print it out.}
         */
        for (nl = result->readdir_res_u.list; nl != NULL; 
          nl = nl->next) @{
                printf("%s\n", nl->name);
        @}
        exit(0);
@}
@end example

Compile everything, and run:

@example
@cartouche
@b{sun%}  rpcgen dir.x
@b{sun%}  cc rls.c dir_clnt.c dir_xdr.c -o rls
@b{sun%}  cc dir_svc.c dir_proc.c dir_xdr.c -o dir_svc

@b{sun%}  dir_svc &
@end cartouche
@cartouche
@b{moon%}  rls sun /usr/pub
.
..
ascii
eqnchar
greek
kbd
marg8
tabclr
tabs
tabs4
@b{moon%}
@end cartouche
@end example

@cindex Debugging with @code{rpcgen}

A final note about @code{rpcgen}: The client program and the server
procedure can be tested together as a single program by simply linking
them with each other rather than with the client and server stubs.  The
procedure calls will be executed as ordinary local procedure calls and
the program can be debugged with a local debugger such as @code{dbx}.
When the program is working, the client program can be linked to the
client stub produced by @code{rpcgen} and the server procedures can be
linked to the server stub produced by @code{rpcgen}.

@strong{NOTE:} @emph{If you do this, you may want to comment out calls
to RPC library routines, and have client-side routines call server
routines directly.}


@node The C-Preprocessor
@section The C-Preprocessor
@iftex
@cindex @code{rpcgen} C-preprocessor
@end iftex
@ifinfo
@cindex rpcgen C-preprocessor
@end ifinfo
@cindex C-preprocessor
@cindex Preprocessor

The C-preprocessor is run on all input files before they are compiled,
so all the preprocessor directives are legal within a @samp{.x} file.
Four symbols may be defined, depending upon which output file is getting
generated. The symbols are:
@table @code
@item @i{Symbol}
@i{Usage}

@item RPC_HDR
for header-file output
@item RPC_XDR
for XDR routine output
@item RPC_SVC
for server-skeleton output
@item RPC_CLNT
for client stub output
@end table

Also, @code{rpcgen} does a little preprocessing of its own. Any line
that begins with a percent sign is passed directly into the output file,
without any interpretation of the line.  Here is a simple example that
demonstrates the preprocessing features.

@example
/*
 * @i{time.x: Remote time protocol}
 */
program TIMEPROG @{
        version TIMEVERS @{
                unsigned int TIMEGET(void) = 1;
        @} = 1;
@} = 44;

#ifdef RPC_SVC
%int *
%timeget_1()
%@{
%        static int thetime;
%
%        thetime = time(0);
%        return (&thetime);
%@}
#endif
@end example

The @samp{%} feature is not generally recommended, as there is no
guarantee that the compiler will stick the output where you intended.

@node rpcgen Programming Notes
@section @code{rpcgen} Programming Notes
@iftex
@cindex @code{rpcgen} other operations
@end iftex
@ifinfo
@cindex rpcgen other operations
@end ifinfo

@menu
* Timeout Changes::
* Handling Broadcast on the Server Side::
* Other Information Passed to Server Procedures::
@end menu

@node Timeout Changes
@subsection Timeout Changes
@cindex Timeout changes

RPC sets a default timeout of 25 seconds for RPC calls when
@code{clnt_create()} is used.  This timeout may be changed using
@code{clnt_control()}. Here is a small code fragment to demonstrate use
of @code{clnt_control()}:

@example
struct timeval tv;
CLIENT *cl;

cl = clnt_create("somehost", SOMEPROG, SOMEVERS, "tcp");
if (cl == NULL) @{
        exit(1);
@}
tv.tv_sec = 60; /* @i{change timeout to 1 minute} */
tv.tv_usec = 0;
clnt_control(cl, CLSET_TIMEOUT, &tv);   
@end example


@node Handling Broadcast on the Server Side
@subsection Handling Broadcast on the Server Side
@cindex Broadcast RPC, server response
@iftex
@cindex @code{rpcgen}, broadcast RPC
@end iftex
@ifinfo
@cindex rpcgen, broadcast RPC
@end ifinfo

When a procedure is known to be called via broadcast RPC, it is usually
wise for the server to not reply unless it can provide some useful
information to the client.  This prevents the network from getting
flooded by useless replies.

To prevent the server from replying, a remote procedure can return
@code{NULL} as its result, and the server code generated by
@code{rpcgen} will detect this and not send out a reply.

Here is an example of a procedure that replies only if it thinks it is
an NFS server:

@example
void *
reply_if_nfsserver()
@{
        char notnull;   /* @i{just here so we can use its address} */

        if (access("/etc/exports", F_OK) < 0) @{
                return (NULL);  /* @i{prevent RPC from replying} */
        @}
        /*
         * @i{return non-null pointer so RPC will send out a reply}
         */
        return ((void *)&notnull);
@}
@end example

Note that if procedure returns type @code{void *}, they must return a
non-@code{NULL} pointer if they want RPC to reply for them.

@node Other Information Passed to Server Procedures
@subsection Other Information Passed to Server Procedures

Server procedures will often want to know more about an RPC call than
just its arguments.  For example, getting authentication information is
important to procedures that want to implement some level of security.
This extra information is actually supplied to the server procedure as a
second argument.  Here is an example to demonstrate its use.  What we've
done here is rewrite the previous @code{printmessage_1()} procedure to
only allow root users to print a message to the console.

@example
int *
printmessage_1(msg, rq)
        char **msg;
        struct svc_req  *rq;
@{
        static in result;       /* @i{Must be static} */
        FILE *f;
        struct authunix_parms *aup;

        aup = (struct authunix_parms *)rq->rq_clntcred;
        if (aup->aup_uid != 0) @{
                result = 0;
                return (&result);
        @}

        /*
         * @i{Same code as before.}
         */
@}
@end example


@node RPC Language
@section RPC Language
@cindex RPC language
@cindex RPCL
@iftex
@cindex @code{rpcgen}, RPC language
@end iftex
@ifinfo
@cindex rpcgen, RPC language
@end ifinfo

RPC language is an extension of XDR language.  The sole extension is the
addition of the @code{program} type.  For a complete description of the
XDR language syntax, @xref{XDR Protocol Specification}.  For a
description of the RPC extensions to the XDR language, @xref{Remote
Procedure Calls Protocol Specification}.

However, XDR language is so close to C that if you know C, you know most
of it already.  We describe here the syntax of the RPC language, showing
a few examples along the way.  We also show how the various RPC and XDR
type definitions get compiled into C type definitions in the output
header file.

@menu
* Definitions::
* Structures::
* Unions::
* Enumerations::
* Typedef::
* Constants::
* Programs::
* Declarations::
* Special Cases::
@end menu

@node Definitions
@subsection Definitions
@cindex Definitions
@iftex
@cindex @code{rpcgen} definitions
@end iftex
@ifinfo
@cindex rpcgen definitions
@end ifinfo
An RPC language file consists of a series of definitions.

@example
    definition-list:
        definition ";"
        definition ";" definition-list
@end example

It recognizes five types of definitions.

@example
    definition:
        enum-definition
        struct-definition
        union-definition
        typedef-definition
        const-definition
        program-definition
@end example

@node Structures
@subsection Structures
@cindex Structures
@iftex
@cindex @code{rpcgen} structures
@end iftex
@ifinfo
@cindex rpcgen structures
@end ifinfo

An XDR struct is declared almost exactly like its C counterpart.  It
looks like the following:

@example
    struct-definition:
        "struct" struct-ident "@{"
            declaration-list
        "@}"

    declaration-list:
        declaration ";"
        declaration ";" declaration-list
@end example

As an example, here is an XDR structure to a two-dimensional coordinate,
and the C structure that it gets compiled into in the output header
file.

@example
   struct coord @{             struct coord @{
        int x;       -->           int x;
        int y;                     int y;
   @};                         @};
                              typedef struct coord coord;
@end example

The output is identical to the input, except for the added
@code{typedef} at the end of the output.  This allows one to use
@code{coord} instead of @code{struct coord} when declaring items.

@node Unions
@subsection Unions
@cindex Unions
@iftex
@cindex @code{rpcgen} unions
@end iftex
@ifinfo
@cindex rpcgen unions
@end ifinfo

XDR unions are discriminated unions, and look quite different from C
unions. They are more analogous to Pascal variant records than they are
to C unions.

@example
    union-definition:
        "union" union-ident "switch" "(" declaration ")" "@{"
            case-list
        "@}"

    case-list:
        "case" value ":" declaration ";"
        "default" ":" declaration ";"
        "case" value ":" declaration ";" case-list
@end example

Here is an example of a type that might be returned as the result of a
``read data'' operation.  If there is no error, return a block of data.
Otherwise, don't return anything.

@example
    union read_result switch (int errno) @{
    case 0:
        opaque data[1024];
    default:
        void;
    @};
@end example

It gets compiled into the following:

@example
    struct read_result @{
        int errno;
        union @{
            char data[1024];
        @} read_result_u;
    @};
    typedef struct read_result read_result;
@end example

Notice that the union component of the output struct has the name as the
type name, except for the trailing @samp{_u}.

@node Enumerations
@subsection Enumerations
@cindex Enumerations
@iftex
@cindex @code{rpcgen} enumerations
@end iftex
@ifinfo
@cindex rpcgen enumerations
@end ifinfo

XDR enumerations have the same syntax as C enumerations.

@example
    enum-definition:
        "enum" enum-ident "@{"
            enum-value-list
        "@}"

    enum-value-list:
        enum-value
        enum-value "," enum-value-list

    enum-value:
        enum-value-ident 
        enum-value-ident "=" value
@end example

Here is a short example of an XDR enum, and the C enum that it gets
compiled into.

@example
     enum colortype @{      enum colortype @{
          RED = 0,              RED = 0,
          GREEN = 1,   -->      GREEN = 1,
          BLUE = 2              BLUE = 2,
     @};                    @};
                           typedef enum colortype colortype;
@end example

@node Typedef
@subsection Typedef
@cindex Typedef
@iftex
@cindex @code{rpcgen} typedef
@end iftex
@ifinfo
@cindex rpcgen typedef
@end ifinfo

XDR typedefs have the same syntax as C typedefs.

@example
    typedef-definition:
        "typedef" declaration
@end example

Here is an example that defines a @code{fname_type} used for declaring
file name strings that have a maximum length of 255 characters.

@example
typedef string fname_type<255>; --> typedef char *fname_type;
@end example

@node Constants
@subsection Constants
@cindex Constants
@iftex
@cindex @code{rpcgen} constants
@end iftex
@ifinfo
@cindex rpcgen constants
@end ifinfo

XDR constants symbolic constants that may be used wherever a integer
constant is used, for example, in array size specifications.

@example
    const-definition:
        "const" const-ident "=" integer
@end example

For example, the following defines a constant @code{DOZEN} equal to 12.

@example
    const DOZEN = 12;  -->  #define DOZEN 12
@end example

@node Programs
@subsection Programs
@cindex Programs
@iftex
@cindex @code{rpcgen} programs
@end iftex
@ifinfo
@cindex rpcgen programs
@end ifinfo

RPC programs are declared using the following syntax:

@example
    program-definition:
        "program" program-ident "@{" 
            version-list
        "@}" "=" value 

    version-list:
        version ";"
        version ";" version-list

    version:
        "version" version-ident "@{"
            procedure-list 
        "@}" "=" value

    procedure-list:
        procedure ";"
        procedure ";" procedure-list

    procedure:
        type-ident procedure-ident "(" type-ident ")" "=" value
@end example

For example, here is the time protocol, revisited:

@example
/*
 * @i{time.x: Get or set the time. Time is represented as number of seconds}
 * @i{since 0:00, January 1, 1970.}
 */
program TIMEPROG @{
    version TIMEVERS @{
        unsigned int TIMEGET(void) = 1;
        void TIMESET(unsigned) = 2;
    @} = 1;
@} = 44;        
@end example

This file compiles into @code{#defines} in the output header file:

@example
#define TIMEPROG 44
#define TIMEVERS 1
#define TIMEGET 1
#define TIMESET 2
@end example

@node Declarations
@subsection Declarations
@cindex Declarations
@iftex
@cindex @code{rpcgen} declarations
@end iftex
@ifinfo
@cindex rpcgen declarations
@end ifinfo

In XDR, there are only four kinds of declarations.

@example
    declaration:
        simple-declaration
        fixed-array-declaration
        variable-array-declaration
        pointer-declaration
@end example

@menu
* Simple Declarations::
* Fixed-Length Array Declarations::
* Variable-Length Array Declarations::
* Pointer Declarations::
@end menu

@node Simple Declarations
@subsubsection Simple Declarations
@cindex Simple declarations

Simple declarations are just like simple C declarations.

@example
    simple-declaration:
        type-ident variable-ident
@end example

Example:

@example
    colortype color;    --> colortype color;
@end example

@node Fixed-Length Array Declarations
@subsubsection Fixed-Length Array Declarations
@cindex Fixed-length array declarations

Fixed-length array declarations are just like C array declarations:

@example
    fixed-array-declaration:
        type-ident variable-ident "[" value "]"
@end example

Example:

@example
    colortype palette[8];    --> colortype palette[8];
@end example

@node Variable-Length Array Declarations
@subsubsection Variable-Length Array Declarations
@cindex Variable-length array declarations

Variable-length array declarations have no explicit syntax in C, so XDR
invents its own using angle-brackets.

@example
variable-array-declaration:
    type-ident variable-ident "<" value ">"
    type-ident variable-ident "<" ">"
@end example

The maximum size is specified between the angle brackets. The size may
be omitted, indicating that the array may be of any size.

@example
    int heights<12>;    /* @i{at most 12 items} */
    int widths<>;       /* @i{any number of items} */
@end example

Since variable-length arrays have no explicit syntax in C, these
declarations are actually compiled into @code{struct}s.  For example, the
@samp{heights} declaration gets compiled into the following struct:

@example
    struct @{
        u_int heights_len;  /* @i{# of items in array} */
        int *heights_val;   /* @i{pointer to array} */
    @} heights;
@end example

Note that the number of items in the array is stored in the @samp{_len}
component and the pointer to the array is stored in the @samp{_val}
component. The first part of each of these component's names is the same
as the name of the declared XDR variable.

@node Pointer Declarations
@subsubsection Pointer Declarations
@cindex Pointer declarations

Pointer declarations are made in XDR exactly as they are in C.  You
can't really send pointers over the network, but you can use XDR
pointers for sending recursive data types such as lists and trees.  The
type is actually called @samp{optional-data}, not @samp{pointer}, in XDR
language.

@example
    pointer-declaration:
        type-ident "*" variable-ident
@end example

Example:

@example
    listitem *next;  -->  listitem *next;
@end example

@node Special Cases
@subsection Special Cases
@iftex
@cindex @code{rpcgen} special cases
@end iftex
@ifinfo
@cindex rpcgen special cases
@end ifinfo

There are a few exceptions to the rules described above.

@menu
* XDR Booleans::
* XDR Strings::
* XDR Opaque Data::
* XDR Voids::
@end menu

@node XDR Booleans
@subsubsection Booleans
@cindex Booleans
@iftex
@cindex @code{bool}
@cindex @code{bool_t}
@end iftex
@ifinfo
@cindex bool
@cindex bool_t
@end ifinfo

C has no built-in boolean type. However, the RPC library does a boolean
type called @code{bool_t} that is either @code{TRUE} or @code{FALSE}.
Things declared as type @code{bool} in XDR language are compiled into
@code{bool_t} in the output header file.

Example:

@example
    bool married;  -->  bool_t married;
@end example

@node XDR Strings
@subsubsection Strings
@cindex XDR Strings
@cindex Strings
@iftex
@cindex @code{string}
@end iftex
@ifinfo
@cindex string
@end ifinfo

C has no built-in string type, but instead uses the null-terminated
@code{char *} convention.  In XDR language, strings are declared using
the @code{string} keyword, and compiled into @code{char *}s in the
output header file. The maximum size contained in the angle brackets
specifies the maximum number of characters allowed in the strings (not
counting the @code{NULL} character). The maximum size may be left off,
indicating a string of arbitrary length.

Examples:

@example
    string name<32>;    -->  char *name;
    string longname<>;  -->  char *longname;
@end example

@node XDR Opaque Data
@subsubsection Opaque Data
@cindex Opaque data
@iftex
@cindex @code{opaque}
@end iftex
@ifinfo
@cindex opaque
@end ifinfo

Opaque data is used in RPC and XDR to describe untyped data, that is,
just sequences of arbitrary bytes.  It may be declared either as a fixed
or variable length array.

Examples:

@example
    opaque diskblock[512];  -->  char diskblock[512];

    opaque filedata<1024>;  -->  struct @{
                                    u_int filedata_len;
                                    char *filedata_val;
                                 @} filedata;
@end example

@node XDR Voids
@subsubsection Voids
@cindex Voids
@iftex
@cindex @code{void}
@end iftex
@ifinfo
@cindex void
@end ifinfo

In a void declaration, the variable is not named.  The declaration is
just @code{void} and nothing else.  Void declarations can only occur in
two places: union definitions and program definitions (as the argument
or result of a remote procedure).

